package account

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"sort"
	"time"

	"github.com/go-redis/redis/v8"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("account")

var scriptLua = redis.NewScript(`
  local cursor = ARGV[1]
  local count = ARGV[2]
  local res = {}
  local n_accounts
  repeat
    local accts = redis.call("hscan", "accounts:", cursor, "COUNT", count)
    cursor = tonumber(accts[1])
    n_accounts = #(accts[2])
    for i, acc in ipairs(accts[2]) do
      if i % 2 == 1 then

        local ok = true -- by default, allow account

        -- KEYS are include patterns
        if #KEYS > 0 then
          ok = false -- by default, reject account when there are include patterns
          for i, r in ipairs(KEYS) do
            if acc:match(r) then
              ok = true -- allow account
              break
            end
          end
        end

        -- ARGV are reject patterns
        if ok then
          for i = 3, #ARGV do
            if acc:match(ARGV[i]) then
              ok = false -- reject account
              break
            end
          end
        end

        if ok then
          local values = redis.call("HMGET", "account:" .. acc, "bytes", "objects")
          if #values == 2 then
            res[acc] = {
              bytes = tonumber(values[1]),
              objects = tonumber(values[2]),
              containers = redis.call("ZCOUNT", "containers:" .. acc, 0, 0),
              buckets = redis.call("ZCARD", "buckets:" .. acc)
            }
          end
        end
      end
    end
  until cursor == 0 or n_accounts > count/2

  return cjson.encode({cursor = cursor, accounts = res})
`)

var confTimeout = 30 * time.Second
var confInterval = 15 * time.Minute
var confIntervalScan = 2 * time.Second
var confRefreshSentinel = 5 * time.Minute
var ConfFiltersAllow common.ArrayFlags
var ConfFiltersReject common.ArrayFlags
var confScanCount = uint(1000)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "account.timeout", confTimeout, "[account] operation timeout")
	fs.DurationVar(&confInterval, "account.interval", confInterval, "[account] interval between polls")
	fs.DurationVar(&confIntervalScan, "account.interval.scan", confIntervalScan, "[account] limited interval between scan polls")
	fs.DurationVar(&confRefreshSentinel, "account.refresh", confRefreshSentinel, "[account] interval between refresh of sentinel state")
	fs.Var(&ConfFiltersAllow, "account.filter.allow", "[account] regex(es) to select which accounts to gather. Default no filtering")
	fs.Var(&ConfFiltersReject, "account.filter.reject", "[account] regex(es) to limit which accounts not to gather. Default no filtering")
	fs.UintVar(&confScanCount, "account.scan.count", confScanCount, "Number of entries for scan() redis function to return")
}

var Factory = &accountDaemonFactory{}

type accountDaemonFactory struct{}

func (df *accountDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})
	namespace := inventory.Namespace

	localRedis := []string{}
	for _, service := range inventory.Services["redis"] {
		localRedis = append(localRedis, fmt.Sprintf("%s:%d", service.IP, service.Port))
	}

	for _, service := range inventory.Services["redissentinel"] {
		params[service.ID] = &accountDaemonParams{
			Namespace:   namespace,
			ServiceID:   service.ID,
			ServiceIP:   service.IP,
			ServicePort: service.Port,
			LocalRedis:  localRedis,
		}
	}

	return params
}

func (df *accountDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*accountDaemonParams)

	return &accountDaemon{
		namespace:   params.Namespace,
		serviceID:   params.ServiceID,
		serviceIP:   params.ServiceIP,
		servicePort: params.ServicePort,
		LocalRedis:  params.LocalRedis,
	}
}

type accountDaemonParams struct {
	Namespace   string
	ServiceID   string
	ServiceIP   string
	ServicePort int
	LocalRedis  []string
}

type accountDaemon struct {
	daemon.BaseDaemon

	namespace   string
	serviceID   string
	serviceIP   string
	servicePort int
	serviceAddr string
	masterName  string
	slaves      []string
	lastUpdate  time.Time
	LocalRedis  []string

	redisClient    *redis.Client
	sentinelClient *redis.SentinelClient
	cursor         uint64
}

func (d *accountDaemon) Interval() time.Duration {
	if d.cursor == 0 {
		return confInterval
	}

	return confIntervalScan
}

func (d *accountDaemon) Init(ctx context.Context) error {
	d.serviceAddr = fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)

	d.sentinelClient = redis.NewSentinelClient(&redis.Options{
		Addr:         d.serviceAddr,
		DialTimeout:  confTimeout,
		ReadTimeout:  confTimeout,
		WriteTimeout: confTimeout,
	}).WithContext(ctx)

	return nil
}

func (d *accountDaemon) updateSentinel(ctx context.Context) error {
	forceUpdate := false
	if d.lastUpdate.IsZero() || time.Since(d.lastUpdate) > confRefreshSentinel {
		forceUpdate = true
		log.Debugf("Force sentinel refresh")
	}

	if d.masterName == "" || forceUpdate {
		masters, err := d.sentinelClient.Masters(ctx).Result()
		if err != nil {
			return err
		}
		if len(masters) < 1 {
			return fmt.Errorf("Sentinel returned no master")
		}
		master := masters[0].([]interface{})
		if len(master) < 2 {
			return fmt.Errorf("Sentinel incomplete master informations")
		}
		d.masterName = master[1].(string)
		log.Debugf("got sentinel master name: %s", d.masterName)
	}

	if d.slaves == nil || forceUpdate {
		slaves, err := d.sentinelClient.Slaves(ctx, d.masterName).Result()
		if err != nil {
			return err
		}

		d.slaves = make([]string, len(slaves))
		for i, v := range slaves {
			slave := v.([]interface{})
			ipPort := slave[3].(string) + ":" + slave[5].(string)
			log.Debugf("slaves %s", ipPort)
			d.slaves[i] = ipPort
		}

		sort.Strings(d.slaves)
		log.Debugf("slaves: %v", d.slaves)
	}

	if forceUpdate {
		d.lastUpdate = time.Now()
		log.Debugf("update lastUpdate after force refresh")
	}

	for _, r := range d.LocalRedis {
		if r == d.slaves[0] {
			continue
		}
		log.Debugf("The first redis slave is owned locally, let's use it")
		if d.redisClient == nil || d.redisClient.Options().Addr != r {
			log.Debugf("Create new redis client to %s", r)
			if d.redisClient != nil {
				d.redisClient.Close()
			}
			d.redisClient = redis.NewClient(&redis.Options{
				Addr:         r,
				DialTimeout:  confTimeout,
				ReadTimeout:  confTimeout,
				WriteTimeout: confTimeout,
			}).WithContext(ctx)
		}
		return nil
	}
	log.Debugf("The first redis slave is not owned locally, let's skip for now")
	return nil
}

func (d *accountDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	log.Debugf("Collecting (%v) ...", d)

	if err := d.updateSentinel(ctx); err != nil {
		return err
	}

	if d.redisClient == nil {
		return nil
	}

	rejects := make([]interface{}, len(ConfFiltersReject)+2)
	rejects[0] = d.cursor
	rejects[1] = confScanCount
	for i, r := range ConfFiltersReject {
		rejects[i+2] = r
	}

	info, err := scriptLua.Run(ctx, d.redisClient, ConfFiltersAllow, rejects).Result()
	if err != nil {
		m.IncrError("account", "account")
		return err
	}
	type jsonUnit struct {
		Bytes, Objects, Containers, Buckets uint64
	}
	type jsonStruct struct {
		Cursor   uint64
		Accounts map[string]jsonUnit
	}
	var obj jsonStruct
	err = json.Unmarshal([]byte(info.(string)), &obj)
	if err != nil {
		m.IncrError("account", "account")
		return err
	}
	d.cursor = obj.Cursor
	for account, data := range obj.Accounts {
		d.sendStat("size_bytes", account, data.Bytes, m)
		d.sendStat("objects", account, data.Objects, m)
		d.sendStat("buckets", account, data.Buckets, m)
		d.sendStat("containers", account, data.Containers, m)
	}
	return nil
}

func (d *accountDaemon) sendStat(metric string, account string, value uint64, m *common.MetricData) {
	labels := map[string]string{
		"namespace": d.namespace,
		"account":   account,
	}
	m.Set(&common.Metric{
		Name:   "account_" + metric,
		Labels: labels,
		Value:  float64(value),
		TTL:    confInterval * 2,
	})
}

func (d *accountDaemon) Close() error {
	if d.redisClient != nil {
		d.redisClient.Close()
	}
	if d.sentinelClient != nil {
		d.sentinelClient.Close()
	}
	return nil
}
