package log

import (
	"flag"
	"fmt"
	"log/syslog"
	"os"
	"path"
	"strings"
)

type Level int

const (
	DEBUG Level = iota
	VERBOSE
	INFO
	NOTICE
	WARNING
	WARN
	ERROR
	ERR
	CRITIC
	CRIT
	ALERT
)

var LogLevel Level = NOTICE
var logLevelName string
var logIdentifier string
var logPriority string
var logNoSyslog bool
var logDebugTrace bool

var levelsToNames = map[Level]string{
	DEBUG:   "debug",
	VERBOSE: "verbose",
	INFO:    "info",
	NOTICE:  "notice",
	WARNING: "warning",
	ERROR:   "error",
	CRITIC:  "critic",
	ALERT:   "alert",
}

type syslogFunc func(string) error

var levelsToSyslog map[Level]syslogFunc

func InitFlags(fs *flag.FlagSet) {
	fs.StringVar(&logLevelName, "log.level", levelsToNames[LogLevel], "Log Level")
	fs.StringVar(&logIdentifier, "log.id", path.Base(os.Args[0]), "Syslog identifer/tag")
	fs.StringVar(&logPriority, "log.priority", "LOCAL7", "Syslog priority")
	fs.BoolVar(&logNoSyslog, "log.nosyslog", false, "Disable syslog logging ?")
	fs.BoolVar(&logDebugTrace, "log.debugtrace", false, "Add a log trace with function, file and line")
}

func Setup() error {
	findLevel := false
	for level, name := range levelsToNames {
		if name == logLevelName {
			LogLevel = level
			findLevel = true
			break
		}
	}
	if !findLevel {
		return fmt.Errorf("unknown log.level %s", logLevelName)
	}

	if logNoSyslog {
		return nil
	}

	var prio syslog.Priority
	switch strings.ToUpper(logPriority) {
	case "KERN":
	case "LOG_KERN":
		prio = syslog.LOG_KERN
	case "USER":
	case "LOG_USER":
		prio = syslog.LOG_USER
	case "MAIL":
	case "LOG_MAIL":
		prio = syslog.LOG_MAIL
	case "DAEMON":
	case "LOG_DAEMON":
		prio = syslog.LOG_DAEMON
	case "AUTH":
	case "LOG_AUTH":
		prio = syslog.LOG_AUTH
	case "SYSLOG":
	case "LOG_SYSLOG":
		prio = syslog.LOG_SYSLOG
	case "LPR":
	case "LOG_LPR":
		prio = syslog.LOG_LPR
	case "NEWS":
	case "LOG_NEWS":
		prio = syslog.LOG_NEWS
	case "UUCP":
	case "LOG_UUCP":
		prio = syslog.LOG_UUCP
	case "CRON":
	case "LOG_CRON":
		prio = syslog.LOG_CRON
	case "AUTHPRIV":
	case "LOG_AUTHPRIV":
		prio = syslog.LOG_AUTHPRIV
	case "FTP":
	case "LOG_FTP":
		prio = syslog.LOG_FTP
	case "LOCAL0":
	case "LOG_LOCAL0":
		prio = syslog.LOG_LOCAL0
	case "LOCAL1":
	case "LOG_LOCAL1":
		prio = syslog.LOG_LOCAL1
	case "LOCAL2":
	case "LOG_LOCAL2":
		prio = syslog.LOG_LOCAL2
	case "LOCAL3":
	case "LOG_LOCAL3":
		prio = syslog.LOG_LOCAL3
	case "LOCAL4":
	case "LOG_LOCAL4":
		prio = syslog.LOG_LOCAL4
	case "LOCAL5":
	case "LOG_LOCAL5":
		prio = syslog.LOG_LOCAL5
	case "LOCAL6":
	case "LOG_LOCAL6":
		prio = syslog.LOG_LOCAL6
	case "LOCAL7":
	case "LOG_LOCAL7":
		prio = syslog.LOG_LOCAL7
	default:
		return fmt.Errorf("unknown log.priority %s", logPriority)
	}

	logger, err := syslog.New(prio, logIdentifier)
	if err != nil {
		return fmt.Errorf("unable to open syslog: %w", err)
	}
	levelsToSyslog = map[Level]syslogFunc{
		DEBUG:   logger.Debug,
		VERBOSE: logger.Debug,
		INFO:    logger.Info,
		NOTICE:  logger.Notice,
		WARNING: logger.Warning,
		ERROR:   logger.Err,
		CRITIC:  logger.Crit,
		ALERT:   logger.Alert,
	}
	return nil
}
